@@ -3,7 +3,7 @@
 """
 DOM wrapper around libxml2, specifically the libxml2mod Python extension module.
 
-Copyright (C) 2003, 2004, 2005, 2006 Paul Boddie <paul@boddie.org.uk>
+Copyright (C) 2003, 2004, 2005, 2006, 2007 Paul Boddie <paul@boddie.org.uk>
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
@@ -20,7 +20,7 @@ License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 """
 
-__version__ = "0.3.6"
+__version__ = "0.4"
 
 from libxml2dom.macrolib import *
 from libxml2dom.macrolib import \
@@ -29,6 +29,38 @@ from libxml2dom.macrolib import \
     parseFile as Node_parseFile, \
     toString as Node_toString, toStream as Node_toStream, \
     toFile as Node_toFile
+import urllib # for parseURI in HTML mode
+
+class Implementation(object):
+
+    "Contains an abstraction over the DOM implementation."
+
+    def createDocumentType(self, localName, publicId, systemId):
+        return DocumentType(localName, publicId, systemId)
+
+    def createDocument(self, namespaceURI, localName, doctype):
+        return Document(Node_createDocument(namespaceURI, localName, doctype), self)
+
+    # Wrapping of documents.
+
+    def adoptDocument(self, node):
+        return Document(node, self)
+
+    # Factory functions.
+
+    def get_node(self, _node, context_node):
+        if Node_nodeType(_node) == context_node.DOCUMENT_NODE:
+            return context_node.ownerDocument
+        elif Node_nodeType(_node) == context_node.ATTRIBUTE_NODE:
+            return Attribute(_node, self, context_node.ownerDocument, context_node)
+        else:
+            return Node(_node, self, context_node.ownerDocument)
+
+    def get_node_or_none(self, _node, context_node):
+        if _node is None:
+            return None
+        else:
+            return self.get_node(_node, context_node)
 
 # Attribute and node list wrappers.
 
@@ -96,7 +128,7 @@ class NamedNodeMap(object):
         pass
 
     def values(self):
-        return [Attribute(_node, self.node.ownerDocument) for _node in Node_attributes(self.node.as_native_node()).values()]
+        return [Attribute(_node, self.impl, self.node.ownerDocument) for _node in Node_attributes(self.node.as_native_node()).values()]
 
     def keys(self):
         return [(attr.namespaceURI, attr.localName) for attr in self.values()]
@@ -146,8 +178,9 @@ class Node(object):
     PROCESSING_INSTRUCTION_NODE = xml.dom.Node.PROCESSING_INSTRUCTION_NODE
     TEXT_NODE = xml.dom.Node.TEXT_NODE
 
-    def __init__(self, node, ownerDocument=None):
+    def __init__(self, node, impl=None, ownerDocument=None):
         self._node = node
+        self.impl = impl or default_impl
         self.ownerDocument = ownerDocument
 
     def as_native_node(self):
@@ -160,7 +193,7 @@ class Node(object):
 
         # NOTE: Consider a generator instead.
 
-        return NodeList([Node(_node, self.ownerDocument) for _node in Node_childNodes(self._node)])
+        return NodeList([self.impl.get_node(_node, self) for _node in Node_childNodes(self._node)])
 
     def _attributes(self):
         return NamedNodeMap(self)
@@ -168,7 +201,12 @@ class Node(object):
     def _namespaceURI(self):
         return Node_namespaceURI(self._node)
 
+    def _textContent(self):
+        return Node_textContent(self._node)
+
     def _nodeValue(self):
+        if self.nodeType in null_value_node_types:
+            return None
         return Node_nodeValue(self._node)
 
     def _setNodeValue(self, value):
@@ -187,16 +225,16 @@ class Node(object):
         return Node_localName(self._node)
 
     def _parentNode(self):
-        return get_node(Node_parentNode(self._node), self)
+        return self.impl.get_node_or_none(Node_parentNode(self._node), self)
 
     def _previousSibling(self):
-        return Node(Node_previousSibling(self._node), self.ownerDocument)
+        return self.impl.get_node_or_none(Node_previousSibling(self._node), self)
 
     def _nextSibling(self):
-        return Node(Node_nextSibling(self._node), self.ownerDocument)
+        return self.impl.get_node_or_none(Node_nextSibling(self._node), self)
 
     def _doctype(self):
-        return Node(Node_doctype(self._node), self.ownerDocument)
+        return self.impl.get_node(Node_doctype(self._node), self)
 
     def _publicId(self):
         # NOTE: To be fixed when the libxml2mod API has been figured out.
@@ -244,10 +282,10 @@ class Node(object):
         return Node_getAttribute(self._node, name)
 
     def getAttributeNodeNS(self, ns, localName):
-        return Attribute(Node_getAttributeNodeNS(self._node, ns, localName), self.ownerDocument, self)
+        return Attribute(Node_getAttributeNodeNS(self._node, ns, localName), self.impl, self.ownerDocument, self)
 
     def getAttributeNode(self, localName):
-        return Attribute(Node_getAttributeNode(self._node, localName), self.ownerDocument, self)
+        return Attribute(Node_getAttributeNode(self._node, localName), self.impl, self.ownerDocument, self)
 
     def setAttributeNS(self, ns, name, value):
         Node_setAttributeNS(self._node, ns, name, value)
@@ -268,30 +306,30 @@ class Node(object):
         Node_removeAttribute(self._node, name)
 
     def createElementNS(self, ns, name):
-        return Node(Node_createElementNS(self._node, ns, name), self.ownerDocument)
+        return self.impl.get_node(Node_createElementNS(self._node, ns, name), self)
 
     def createElement(self, name):
-        return Node(Node_createElement(self._node, name), self.ownerDocument)
+        return self.impl.get_node(Node_createElement(self._node, name), self)
 
     def createAttributeNS(self, ns, name):
         tmp = self.createElement("tmp")
-        return Attribute(Node_createAttributeNS(tmp._node, ns, name))
+        return Attribute(Node_createAttributeNS(tmp._node, self.impl, ns, name))
 
     def createAttribute(self, name):
         tmp = self.createElement("tmp")
-        return Attribute(Node_createAttribute(tmp._node, name))
+        return Attribute(Node_createAttribute(tmp._node, name), self.impl)
 
     def createTextNode(self, value):
-        return Node(Node_createTextNode(self._node, value), self.ownerDocument)
+        return self.impl.get_node(Node_createTextNode(self._node, value), self)
 
     def createComment(self, value):
-        return Node(Node_createComment(self._node, value), self.ownerDocument)
+        return self.impl.get_node(Node_createComment(self._node, value), self)
 
     def importNode(self, node, deep):
         if hasattr(node, "as_native_node"):
-            return Node(Node_importNode(self._node, node.as_native_node(), deep), self.ownerDocument)
+            return self.impl.get_node(Node_importNode(self._node, node.as_native_node(), deep), self)
         else:
-            return Node(Node_importNode_DOM(self._node, node, deep), self.ownerDocument)
+            return self.impl.get_node(Node_importNode_DOM(self._node, node, deep), self)
 
     def cloneNode(self, deep):
         # This takes advantage of the ubiquity of importNode (in spite of the DOM specification).
@@ -299,21 +337,21 @@ class Node(object):
 
     def insertBefore(self, tmp, oldNode):
         if hasattr(tmp, "as_native_node"):
-            return Node(Node_insertBefore(self._node, tmp.as_native_node(), oldNode.as_native_node()), self.ownerDocument)
+            return self.impl.get_node(Node_insertBefore(self._node, tmp.as_native_node(), oldNode.as_native_node()), self)
         else:
-            return Node(Node_insertBefore(self._node, tmp, oldNode.as_native_node()), self.ownerDocument)
+            return self.impl.get_node(Node_insertBefore(self._node, tmp, oldNode.as_native_node()), self)
 
     def replaceChild(self, tmp, oldNode):
         if hasattr(tmp, "as_native_node"):
-            return Node(Node_replaceChild(self._node, tmp.as_native_node(), oldNode.as_native_node()), self.ownerDocument)
+            return self.impl.get_node(Node_replaceChild(self._node, tmp.as_native_node(), oldNode.as_native_node()), self)
         else:
-            return Node(Node_replaceChild(self._node, tmp, oldNode.as_native_node()), self.ownerDocument)
+            return self.impl.get_node(Node_replaceChild(self._node, tmp, oldNode.as_native_node()), self)
 
     def appendChild(self, tmp):
         if hasattr(tmp, "as_native_node"):
-            return Node(Node_appendChild(self._node, tmp.as_native_node()), self.ownerDocument)
+            return self.impl.get_node(Node_appendChild(self._node, tmp.as_native_node()), self)
         else:
-            return Node(Node_appendChild(self._node, tmp), self.ownerDocument)
+            return self.impl.get_node(Node_appendChild(self._node, tmp), self)
 
     def removeChild(self, tmp):
         if hasattr(tmp, "as_native_node"):
@@ -322,10 +360,10 @@ class Node(object):
             Node_removeChild(self._node, tmp)
 
     def getElementsByTagName(self, tagName):
-        return self.xpath("//" + tagName)
+        return self.xpath(".//" + tagName)
 
     def getElementsByTagNameNS(self, namespaceURI, localName):
-        return self.xpath("//ns:" + localName, namespaces={"ns" : namespaceURI})
+        return self.xpath(".//ns:" + localName, namespaces={"ns" : namespaceURI})
 
     def normalize(self):
         text_nodes = []
@@ -348,6 +386,7 @@ class Node(object):
 
     childNodes = property(_childNodes)
     value = data = nodeValue = property(_nodeValue, _setNodeValue)
+    textContent = property(_textContent)
     name = nodeName = property(_nodeName)
     tagName = property(_tagName)
     namespaceURI = property(_namespaceURI)
@@ -367,11 +406,14 @@ class Node(object):
     entities = {}
     notations = {}
 
-    #def isSameNode(self, other):
-    #    return self._node.nodePath() == other._node.nodePath()
+    def isSameNode(self, other):
+        return self == other
 
-    #def __eq__(self, other):
-    #    return self._node.nodePath() == other._node.nodePath()
+    def __eq__(self, other):
+        return isinstance(other, Node) and libxml2mod.xmlXPathCmpNodes(self._node, other._node) == 0
+
+    def __ne__(self, other):
+        return not (self == other)
 
     # 4DOM extensions to the usual PyXML API.
     # NOTE: To be finished.
@@ -381,7 +423,7 @@ class Node(object):
         if isinstance(result, str):
             return to_unicode(result)
         elif hasattr(result, "__len__"):
-            return NodeList([get_node(_node, self) for _node in result])
+            return NodeList([self.impl.get_node(_node, self) for _node in result])
         else:
             return result
 
@@ -402,8 +444,8 @@ class Attribute(Node):
 
     "A class providing attribute access."
 
-    def __init__(self, node, ownerDocument=None, ownerElement=None):
-        Node.__init__(self, node, ownerDocument)
+    def __init__(self, node, impl, ownerDocument=None, ownerElement=None):
+        Node.__init__(self, node, impl, ownerDocument)
         self.ownerElement = ownerElement
 
     def _parentNode(self):
@@ -417,21 +459,18 @@ class Document(Node):
 
     "A class providing document-level housekeeping."
 
-    def __init__(self, node):
+    def __init__(self, node, impl):
         self._node = node
+        self.impl = impl
 
     def _ownerDocument(self):
         return self
 
-    def _parentNode(self):
-        return None
-
     def __del__(self):
         #print "Freeing document", self._node
         libxml2mod.xmlFreeDoc(self._node)
 
     ownerDocument = property(_ownerDocument)
-    parentNode = property(_parentNode)
 
 class DocumentType(object):
 
@@ -448,82 +487,104 @@ class DocumentType(object):
         self.entities = {}
         self.notations = {}
 
-# Factory functions.
+# Constants.
 
-def get_node(_node, context_node):
-    if Node_nodeType(_node) == context_node.DOCUMENT_NODE:
-        return context_node.ownerDocument
-    elif Node_nodeType(_node) == context_node.ATTRIBUTE_NODE:
-        return Attribute(_node, context_node.ownerDocument, context_node)
-    else:
-        return Node(_node, context_node.ownerDocument)
+null_value_node_types = [
+    Node.DOCUMENT_NODE, Node.DOCUMENT_TYPE_NODE, Node.ELEMENT_NODE,
+    Node.ENTITY_NODE, Node.ENTITY_REFERENCE_NODE, Node.NOTATION_NODE
+    ]
 
 # Utility functions.
 
 def createDocumentType(localName, publicId, systemId):
-    return DocumentType(localName, publicId, systemId)
+    return default_impl.createDocumentType(localName, publicId, systemId)
 
 def createDocument(namespaceURI, localName, doctype):
-    return Document(Node_createDocument(namespaceURI, localName, doctype))
+    return default_impl.createDocument(namespaceURI, localName, doctype)
 
-def parse(stream_or_string, html=0):
+def parse(stream_or_string, html=0, htmlencoding=None, impl=None):
 
     """
     Parse the given 'stream_or_string', where the supplied object can either be
     a stream (such as a file or stream object), or a string (containing the
     filename of a document). If the optional 'html' parameter is set to a true
     value, the content to be parsed will be treated as being HTML rather than
-    XML.
+    XML. If the optional 'htmlencoding' is specified, HTML parsing will be
+    performed with the document encoding assumed to that specified.
 
     A document object is returned by this function.
     """
 
+    impl = impl or default_impl
+
     if hasattr(stream_or_string, "read"):
         stream = stream_or_string
-        return parseString(stream.read(), html)
+        return parseString(stream.read(), html, htmlencoding, impl)
     else:
-        return parseFile(stream_or_string, html)
+        return parseFile(stream_or_string, html, htmlencoding, impl)
 
-def parseFile(filename, html=0):
+def parseFile(filename, html=0, htmlencoding=None, impl=None):
 
     """
     Parse the file having the given 'filename'. If the optional 'html' parameter
     is set to a true value, the content to be parsed will be treated as being
-    HTML rather than XML.
+    HTML rather than XML. If the optional 'htmlencoding' is specified, HTML
+    parsing will be performed with the document encoding assumed to be that
+    specified.
 
     A document object is returned by this function.
     """
 
-    return Document(Node_parseFile(filename, html))
+    impl = impl or default_impl
+    return impl.adoptDocument(Node_parseFile(filename, html, htmlencoding))
 
-def parseString(s, html=0):
+def parseString(s, html=0, htmlencoding=None, impl=None):
 
     """
     Parse the content of the given string 's'. If the optional 'html' parameter
     is set to a true value, the content to be parsed will be treated as being
-    HTML rather than XML.
+    HTML rather than XML. If the optional 'htmlencoding' is specified, HTML
+    parsing will be performed with the document encoding assumed to be that
+    specified.
 
     A document object is returned by this function.
     """
 
-    return Document(Node_parseString(s, html))
+    impl = impl or default_impl
+    return impl.adoptDocument(Node_parseString(s, html, htmlencoding))
 
-def parseURI(uri, html=0):
+def parseURI(uri, html=0, htmlencoding=None, impl=None):
 
     """
     Parse the content found at the given 'uri'. If the optional 'html' parameter
     is set to a true value, the content to be parsed will be treated as being
-    HTML rather than XML.
+    HTML rather than XML. If the optional 'htmlencoding' is specified, HTML
+    parsing will be performed with the document encoding assumed to be that
+    specified.
 
-    The parseURI does not currently work with HTML. Use parse with a stream
-    object instead. For example:
+    XML documents are retrieved using libxml2's own network capabilities; HTML
+    documents are retrieved using the urllib module provided by Python. To
+    retrieve either kind of document using Python's own modules for this purpose
+    (such as urllib), open a stream and pass it to the parse function:
 
-    d = parse(urllib.urlopen("http://www.python.org"), html=1)
+    f = urllib.urlopen(uri)
+    try:
+        doc = libxml2dom.parse(f, html)
+    finally:
+        f.close()
 
     A document object is returned by this function.
     """
 
-    return Document(Node_parseURI(uri, html))
+    if html:
+        f = urllib.urlopen(uri)
+        try:
+            return parse(f, html, htmlencoding, impl)
+        finally:
+            f.close()
+    else:
+        impl = impl or default_impl
+        return impl.adoptDocument(Node_parseURI(uri, html, htmlencoding))
 
 def toString(node, encoding=None, prettyprint=0):
 
@@ -561,7 +622,7 @@ def toFile(node, filename, encoding=None, prettyprint=0):
 
     Node_toFile(node.as_native_node(), filename, encoding, prettyprint)
 
-def adoptNodes(nodes):
+def adoptNodes(nodes, impl=None):
 
     """
     A special utility method which adopts the given low-level 'nodes' and which
@@ -569,12 +630,24 @@ def adoptNodes(nodes):
     should not be casually used.
     """
 
+    impl = impl or default_impl
+
     if len(nodes) == 0:
         return []
-    doc = Document(libxml2mod.doc(nodes[0]))
+    doc = impl.adoptDocument(libxml2mod.doc(nodes[0]))
     results = []
     for node in nodes:
-        results.append(Node(node, doc))
+        results.append(Node(node, impl, doc))
     return results
 
+def getDOMImplementation():
+
+    "Return the default DOM implementation."
+
+    return default_impl
+
+# Single instance of the implementation.
+
+default_impl = Implementation()
+
 # vim: tabstop=4 expandtab shiftwidth=4
