@@ -3,7 +3,7 @@
 """
 DOM macros for virtual libxml2mod node methods and properties.
 
-Copyright (C) 2003, 2004, 2005, 2006 Paul Boddie <paul@boddie.org.uk>
+Copyright (C) 2003, 2004, 2005, 2006, 2007 Paul Boddie <paul@boddie.org.uk>
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
@@ -49,6 +49,10 @@ def to_unicode(s):
     else:
         return s
 
+def check_document(node, tmp):
+    if Node_ownerDocument(tmp) is not None and Node_ownerDocument(node) is not Node_ownerDocument(tmp):
+        raise xml.dom.DOMException(xml.dom.WRONG_DOCUMENT_ERR)
+
 def get_ns(ns):
     out_ns = to_unicode(libxml2mod.xmlNodeGetContent(ns))
     # Detect "" and produce None as the empty namespace.
@@ -140,6 +144,7 @@ def _get_invented_prefix(node, ns):
 
 _nodeTypes = {
     "attribute" : xml.dom.Node.ATTRIBUTE_NODE,
+    "cdata" : xml.dom.Node.CDATA_SECTION_NODE,
     "comment" : xml.dom.Node.COMMENT_NODE,
     "document_xml" : xml.dom.Node.DOCUMENT_NODE,
     "document_html" : xml.dom.Node.DOCUMENT_NODE,
@@ -158,7 +163,7 @@ for label, value in _nodeTypes.items():
     _reverseNodeTypes[value] = label
 
 def Node_ownerDocument(node):
-    return libxml2mod.doc(node) or node
+    return libxml2mod.doc(node)
 
 def Node_nodeType(node):
     return _nodeTypes[libxml2mod.type(node)]
@@ -218,10 +223,13 @@ def Node_setNodeValue(node, value):
     # NOTE: Cannot set attribute node values.
     libxml2mod.xmlNodeSetContent(node, from_unicode(value))
 
-# NOTE: Verify this.
+# NOTE: Verify this. The data attribute should only really exist for text,
+# NOTE: character data, processing instructions and comments.
 
 Node_data = Node_nodeValue
 
+Node_textContent = Node_nodeValue
+
 def Node_prefix(node):
     ns = libxml2mod.xmlNodeGetNs(node)
     if ns is not None:
@@ -404,12 +412,15 @@ def Node_createComment(node, value):
     return libxml2mod.xmlNewComment(value)
 
 def Node_insertBefore(node, tmp, oldNode):
+    check_document(node, tmp)
     return libxml2mod.xmlAddPrevSibling(oldNode, tmp)
 
 def Node_replaceChild(node, tmp, oldNode):
+    check_document(node, tmp)
     return libxml2mod.xmlReplaceNode(oldNode, tmp)
 
 def Node_appendChild(node, tmp):
+    check_document(node, tmp)
     return libxml2mod.xmlAddChild(node, tmp)
 
 def Node_removeChild(node, child):
@@ -435,7 +446,7 @@ def Node_importNode(node, other, deep):
     elif Node_nodeType(other) == xml.dom.Node.COMMENT_NODE:
         return Node_createComment(node, Node_data(other))
 
-    raise ValueError, "Node type '%s' (%d) not supported." % (other, Node_nodeType(other))
+    raise xml.dom.DOMException(xml.dom.NOT_SUPPORTED_ERR, "Node type '%s' (%d) not supported." % (other, Node_nodeType(other)))
 
 def Node_importNode_DOM(node, other, deep):
     if other.nodeType == xml.dom.Node.ELEMENT_NODE:
@@ -457,12 +468,12 @@ def Node_importNode_DOM(node, other, deep):
     elif other.nodeType == xml.dom.Node.COMMENT_NODE:
         return Node_createComment(node, other.data)
 
-    raise ValueError, "Node type '%s' (%d) not supported." % (_reverseNodeTypes[other.nodeType], other.nodeType)
+    raise xml.dom.DOMException(xml.dom.NOT_SUPPORTED_ERR, "Node type '%s' (%d) not supported." % (_reverseNodeTypes[other.nodeType], other.nodeType))
 
 def Node_xpath(node, expr, variables=None, namespaces=None):
     expr = from_unicode(expr)
 
-    context = libxml2mod.xmlXPathNewContext(Node_ownerDocument(node))
+    context = libxml2mod.xmlXPathNewContext(Node_ownerDocument(node) or node)
     libxml2mod.xmlXPathSetContextNode(context, node)
     # NOTE: Discover namespaces from the node.
     # NOTE: Work out how to specify paths without having to use prefixes on
@@ -490,14 +501,14 @@ def createDocument(namespaceURI, localName, doctype):
         libxml2mod.xmlCreateIntSubset(d, doctype.localName, doctype.publicId, doctype.systemId)
     return d
 
-def parse(stream_or_string, html=0):
+def parse(stream_or_string, html=0, htmlencoding=None):
     if hasattr(stream_or_string, "read"):
         stream = stream_or_string
-        return parseString(stream.read(), html)
+        return parseString(stream.read(), html, htmlencoding)
     else:
-        return parseFile(stream_or_string, html)
+        return parseFile(stream_or_string, html, htmlencoding)
 
-def parseFile(s, html=0):
+def parseFile(s, html=0, htmlencoding=None):
     # NOTE: Switching off validation and remote DTD resolution.
     if not html:
         context = libxml2mod.xmlCreateFileParserCtxt(s)
@@ -507,9 +518,9 @@ def parseFile(s, html=0):
         libxml2mod.xmlParseDocument(context)
         return libxml2mod.xmlParserGetDoc(context)
     else:
-        return libxml2mod.htmlReadFile(s, None, HTML_PARSE_NOERROR | HTML_PARSE_NOWARNING | HTML_PARSE_NONET)
+        return libxml2mod.htmlReadFile(s, htmlencoding, HTML_PARSE_NOERROR | HTML_PARSE_NOWARNING | HTML_PARSE_NONET)
 
-def parseString(s, html=0):
+def parseString(s, html=0, htmlencoding=None):
     # NOTE: Switching off validation and remote DTD resolution.
     if not html:
         context = libxml2mod.xmlCreateMemoryParserCtxt(s, len(s))
@@ -521,10 +532,10 @@ def parseString(s, html=0):
     else:
         # NOTE: URL given as None.
         html_url = None
-        return libxml2mod.htmlReadMemory(s, len(s), html_url, None,
+        return libxml2mod.htmlReadMemory(s, len(s), html_url, htmlencoding,
             HTML_PARSE_NOERROR | HTML_PARSE_NOWARNING | HTML_PARSE_NONET)
 
-def parseURI(uri, html=0):
+def parseURI(uri, html=0, htmlencoding=None):
     # NOTE: Switching off validation and remote DTD resolution.
     if not html:
         context = libxml2mod.xmlCreateURLParserCtxt(uri, 0)
